World w;
int selectionIndex = -1;

color solvedColor = color(180, 80, 130);
color backgroundColor = color(240);
color voronoiColor = color(255);
color startColor = color(255, 0, 0);
color goalColor = color (0, 0, 255);
color obstacleColor = color(0);
void setup() {
  //size(1280, 640, "processing.core.PGraphicsRetina2D");
  size(1280, 640);
  w = new World(50);
}

void draw() {
  background(backgroundColor);
  w.update();
  drawGraph();
  drawInfo();
}

void drawGraph() {
  //draw voronoi regions
  noFill();
  stroke(voronoiColor);
  strokeWeight(1);
  for (int i=0; i < w.regions.length; i++) {
    w.regions[i].draw(this);
  }

  textSize(8);
  //draw delaunay edges
  colorMode(HSB, 360, 256, 256);
  strokeWeight(1);
  for (int i=0; i < w.delaunayLinks.length; i++) {
    Node start = w.getNode(w.delaunayLinks[i][0]);
    Node end = w.getNode(w.delaunayLinks[i][1]);
    float edgeMag = start.position.dist(end.position);
    stroke(210, map(edgeMag, 0, 480, 0, 255), map(edgeMag, 0, width, 128, 255));
    if (start.nodeType != NodeType.OBSTACLE || start.nodeType != NodeType.OBSTACLE) {
      if (end.nodeType != NodeType.OBSTACLE || end.nodeType != NodeType.OBSTACLE ) {
        line(start.position.x, start.position.y, end.position.x, end.position.y);
      }
    }
  }

  colorMode(RGB, 255, 255, 255);

  //draw route
  stroke(solvedColor);
  strokeWeight(3);
  for (int i=0; i < w.nodeSize (); i++) {
    Node n = w.getNode(i);
    Node prev = n.prev;
    if (n.isRoute && prev != null) {
      line(n.position.x, n.position.y, prev.position.x, prev.position.y);
    }
  }


  //draw points
  strokeWeight(1);
  noStroke();
  textSize(12);
  for (int i=0; i < w.nodeSize (); i++) {
    Node n = w.getNode(i);
    int r = 5;
    if (n.nodeType == NodeType.START) {
      fill(255, 0, 0, 50);
      ellipse(n.position.x, n.position.y, 30, 30);
      fill(startColor);
      r = 15;
    } else if (n.nodeType == NodeType.GOAL) {
      fill(0, 0, 255, 50);
      ellipse(n.position.x, n.position.y, 30, 30);
      fill(goalColor);
      r = 15;
    } else if (n.nodeType == NodeType.OBSTACLE) {
      fill(0, 50);
      ellipse(n.position.x, n.position.y, 30, 30);
      fill(obstacleColor);
      r = 15;
    } else if (n.isRoute) {
      fill(solvedColor);
      r = 10;
    } else {
      fill(0);
    }
    ellipse(n.position.x, n.position.y, r, r);
    //display index
    text(i, n.position.x+8, n.position.y-3);
    if (n.nodeType == NodeType.START) {
      text("Start", n.position.x+8, n.position.y+10);
    } else if (n.nodeType == NodeType.GOAL) {
      text("Goal", n.position.x+8, n.position.y+10);
    } else if (n.nodeType == NodeType.OBSTACLE) {
      text("Obstacle", n.position.x+8, n.position.y+10);
    }
  }



  // selected point highlighting
  strokeWeight(1);
  if (selectionIndex >= 0) {
    noFill();
    stroke(0);
    ellipse(w.getNode(selectionIndex).position.x, w.getNode(selectionIndex).position.y, 20, 20);


    stroke(200);
    strokeWeight(1);
    textSize(12);
    rectMode(CENTER);

    int[] linkNodes = w.delaunay.getLinked(selectionIndex);
    for (int i=0; i<linkNodes.length; i++) {
      noFill();
      stroke(0);
      //highlight linked node
      ellipse(w.getNode(linkNodes[i]).position.x, w.getNode(linkNodes[i]).position.y, 10, 10);
      if (w.getNode(selectionIndex).nodeType != NodeType.OBSTACLE) {
        //cost
        PVector selecedNodePos = w.getNode(selectionIndex).position;
        PVector linkedNodePos = w.getNode(linkNodes[i]).position;

        PVector edgeCenter = new PVector((selecedNodePos.x + linkedNodePos.x)/2.0, (selecedNodePos.y + linkedNodePos.y)/2.0);
        float cost = dist(selecedNodePos.x, selecedNodePos.y, linkedNodePos.x, linkedNodePos.y);
        fill(128);
        rect(edgeCenter.x, edgeCenter.y, 6, 6);
        text((int)cost, edgeCenter.x + 4, edgeCenter.y + 4);
      }
    }
  }
}

void mousePressed() {
  if (mouseButton == LEFT) {
    int count = 0;
    for (int i=0; i < w.nodeSize (); i++) {
      float distance = dist(w.getNode(i).position.x, w.getNode(i).position.y, mouseX, mouseY);
      if (distance < 10.0) {
        selectionIndex = i;
        count ++;
      }
    }
    if (count == 0) {
      selectionIndex = -1;
    }
  } else if (mouseButton == RIGHT) {
    w.addNode(new Node(mouseX, mouseY));
    selectionIndex = w.nodeSize() -1;
  } else if (mouseButton == CENTER) {
    for (int i=0; i < w.nodes.size (); i++) {
      float distance = dist(w.getNode(i).position.x, w.getNode(i).position.y, mouseX, mouseY);
      if (distance < 10.0) {
        w.removeNode(i);
        selectionIndex = -1;
      }
    }
  }
}

void mouseDragged() {
  if (mouseButton == LEFT || mouseButton == RIGHT) {
    if (selectionIndex >= 0) {
      PVector m = new PVector(mouseX, mouseY);
      if (m.x < 0) {
        m.x = 0;
      } else if (m.x > width) {
        m.x = width;
      }
      if (m.y < 0) {
        m.y = 0;
      } else if (m.y > height) {
        m.y = height;
      }
      w.nodes.get(selectionIndex).position.x = m.x;
      w.nodes.get(selectionIndex).position.y = m.y;
    }
  }
}

void keyReleased() {
  switch(key) {
  case 's':
    w.setStartNode(selectionIndex);
    break;
  case 'g':
    w.setGoalNode(selectionIndex);
    break;
  case 'o':
    w.setObstacleNode(selectionIndex);
    break;
  case ' ':
    selectionIndex = -1;
    break;
  }
}

void drawInfo() {
  fill(0);
  textSize(12);
  text("Waypoint base A* Path Finding", 10, 16);
  text("LMB:Select & drag node/ RMB: Add node/ MMB: Remove Node", 10, height - 18 -5);
  text("Key s: Set selection node to start node/ Key g: Set selection node to goal node / Key o: Toggle Obstacle node", 10, height-5);
}

